"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SvelteSnapshotManager = exports.SvelteSnapshot = void 0;
const svelte2tsx_1 = __importDefault(require("svelte2tsx"));
const source_mapper_1 = require("./source-mapper");
const utils_1 = require("./utils");
class SvelteSnapshot {
    constructor(typescript, fileName, svelteCode, mapper, logger, isTsFile) {
        this.typescript = typescript;
        this.fileName = fileName;
        this.svelteCode = svelteCode;
        this.mapper = mapper;
        this.logger = logger;
        this.isTsFile = isTsFile;
        this.convertInternalCodePositions = false;
    }
    update(svelteCode, mapper) {
        this.svelteCode = svelteCode;
        this.mapper = mapper;
        this.lineOffsets = undefined;
        this.log('Updated Snapshot');
    }
    getOriginalTextSpan(textSpan) {
        if (!utils_1.isNoTextSpanInGeneratedCode(this.getText(), textSpan)) {
            return null;
        }
        const start = this.getOriginalOffset(textSpan.start);
        if (start === -1) {
            return null;
        }
        // Assumption: We don't change identifiers itself, so we don't change ranges.
        return {
            start,
            length: textSpan.length
        };
    }
    getOriginalOffset(generatedOffset) {
        if (!this.scriptInfo) {
            return generatedOffset;
        }
        this.toggleMappingMode(true);
        const lineOffset = this.scriptInfo.positionToLineOffset(generatedOffset);
        this.debug('try convert offset', generatedOffset, '/', lineOffset);
        const original = this.mapper.getOriginalPosition({
            line: lineOffset.line - 1,
            character: lineOffset.offset - 1
        });
        this.toggleMappingMode(false);
        if (original.line === -1) {
            return -1;
        }
        const originalOffset = this.scriptInfo.lineOffsetToPosition(original.line + 1, original.character + 1);
        this.debug('converted offset to', original, '/', originalOffset);
        return originalOffset;
    }
    setAndPatchScriptInfo(scriptInfo) {
        // @ts-expect-error
        scriptInfo.scriptKind = this.typescript.ScriptKind.TSX;
        const positionToLineOffset = scriptInfo.positionToLineOffset.bind(scriptInfo);
        scriptInfo.positionToLineOffset = (position) => {
            if (this.convertInternalCodePositions) {
                const lineOffset = positionToLineOffset(position);
                this.debug('positionToLineOffset for generated code', position, lineOffset);
                return lineOffset;
            }
            const lineOffset = this.positionAt(position);
            this.debug('positionToLineOffset for original code', position, lineOffset);
            return { line: lineOffset.line + 1, offset: lineOffset.character + 1 };
        };
        const lineOffsetToPosition = scriptInfo.lineOffsetToPosition.bind(scriptInfo);
        scriptInfo.lineOffsetToPosition = (line, offset) => {
            if (this.convertInternalCodePositions) {
                const position = lineOffsetToPosition(line, offset);
                this.debug('lineOffsetToPosition for generated code', { line, offset }, position);
                return position;
            }
            const position = this.offsetAt({ line: line - 1, character: offset - 1 });
            this.debug('lineOffsetToPosition for original code', { line, offset }, position);
            return position;
        };
        // TODO do we need to patch this?
        // const lineToTextSpan = scriptInfo.lineToTextSpan.bind(scriptInfo);
        // scriptInfo.lineToTextSpan = (line) => {
        //     if (this.convertInternalCodePositions) {
        //         const span = lineToTextSpan(line);
        //         this.debug('lineToTextSpan for generated code', line, span);
        //         return span;
        //     }
        //     const lineOffset = this.getLineOffsets();
        //     const start = lineOffset[line - 1];
        //     const span: ts.TextSpan = {
        //         start,
        //         length: (lineOffset[line] || this.svelteCode.length) - start
        //     };
        //     this.debug('lineToTextSpan for original code', line, span);
        //     return span;
        // };
        this.scriptInfo = scriptInfo;
        this.log('patched scriptInfo');
    }
    /**
     * Get the line and character based on the offset
     * @param offset The index of the position
     */
    positionAt(offset) {
        offset = this.clamp(offset, 0, this.svelteCode.length);
        const lineOffsets = this.getLineOffsets();
        let low = 0;
        let high = lineOffsets.length;
        if (high === 0) {
            return { line: 0, character: offset };
        }
        while (low < high) {
            const mid = Math.floor((low + high) / 2);
            if (lineOffsets[mid] > offset) {
                high = mid;
            }
            else {
                low = mid + 1;
            }
        }
        // low is the least x for which the line offset is larger than the current offset
        // or array.length if no line offset is larger than the current offset
        const line = low - 1;
        return { line, character: offset - lineOffsets[line] };
    }
    /**
     * Get the index of the line and character position
     * @param position Line and character position
     */
    offsetAt(position) {
        const lineOffsets = this.getLineOffsets();
        if (position.line >= lineOffsets.length) {
            return this.svelteCode.length;
        }
        else if (position.line < 0) {
            return 0;
        }
        const lineOffset = lineOffsets[position.line];
        const nextLineOffset = position.line + 1 < lineOffsets.length
            ? lineOffsets[position.line + 1]
            : this.svelteCode.length;
        return this.clamp(nextLineOffset, lineOffset, lineOffset + position.character);
    }
    getLineOffsets() {
        if (this.lineOffsets) {
            return this.lineOffsets;
        }
        const lineOffsets = [];
        const text = this.svelteCode;
        let isLineStart = true;
        for (let i = 0; i < text.length; i++) {
            if (isLineStart) {
                lineOffsets.push(i);
                isLineStart = false;
            }
            const ch = text.charAt(i);
            isLineStart = ch === '\r' || ch === '\n';
            if (ch === '\r' && i + 1 < text.length && text.charAt(i + 1) === '\n') {
                i++;
            }
        }
        if (isLineStart && text.length > 0) {
            lineOffsets.push(text.length);
        }
        this.lineOffsets = lineOffsets;
        return lineOffsets;
    }
    clamp(num, min, max) {
        return Math.max(min, Math.min(max, num));
    }
    log(...args) {
        this.logger.log('SvelteSnapshot:', this.fileName, '-', ...args);
    }
    debug(...args) {
        this.logger.debug('SvelteSnapshot:', this.fileName, '-', ...args);
    }
    toggleMappingMode(convertInternalCodePositions) {
        this.convertInternalCodePositions = convertInternalCodePositions;
    }
    getText() {
        var _a;
        const snapshot = (_a = this.scriptInfo) === null || _a === void 0 ? void 0 : _a.getSnapshot();
        if (!snapshot) {
            return '';
        }
        return snapshot.getText(0, snapshot.getLength());
    }
}
exports.SvelteSnapshot = SvelteSnapshot;
class SvelteSnapshotManager {
    constructor(typescript, projectService, logger, strictMode) {
        this.typescript = typescript;
        this.projectService = projectService;
        this.logger = logger;
        this.strictMode = strictMode;
        this.snapshots = new Map();
        this.patchProjectServiceReadFile();
    }
    get(fileName) {
        return this.snapshots.get(fileName);
    }
    create(fileName) {
        if (this.snapshots.has(fileName)) {
            return this.snapshots.get(fileName);
        }
        // This will trigger projectService.host.readFile which is patched below
        const scriptInfo = this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(fileName), false);
        if (!scriptInfo) {
            this.logger.log('Was not able get snapshot for', fileName);
            return;
        }
        try {
            scriptInfo.getSnapshot(); // needed to trigger readFile
        }
        catch (e) {
            this.logger.log('Loading Snapshot failed', fileName);
        }
        const snapshot = this.snapshots.get(fileName);
        if (!snapshot) {
            this.logger.log('Svelte snapshot was not found after trying to load script snapshot for', fileName);
            return; // should never get here
        }
        snapshot.setAndPatchScriptInfo(scriptInfo);
        this.snapshots.set(fileName, snapshot);
        return snapshot;
    }
    patchProjectServiceReadFile() {
        const readFile = this.projectService.host.readFile;
        this.projectService.host.readFile = (path) => {
            if (utils_1.isSvelteFilePath(path)) {
                this.logger.debug('Read Svelte file:', path);
                const svelteCode = readFile(path) || '';
                try {
                    const isTsFile = true; // TODO check file contents? TS might be okay with importing ts into js.
                    const result = svelte2tsx_1.default(svelteCode, {
                        filename: path.split('/').pop(),
                        strictMode: this.strictMode,
                        isTsFile
                    });
                    const existingSnapshot = this.snapshots.get(path);
                    if (existingSnapshot) {
                        existingSnapshot.update(svelteCode, new source_mapper_1.SourceMapper(result.map.mappings));
                    }
                    else {
                        this.snapshots.set(path, new SvelteSnapshot(this.typescript, path, svelteCode, new source_mapper_1.SourceMapper(result.map.mappings), this.logger, isTsFile));
                    }
                    this.logger.log('Successfully read Svelte file contents of', path);
                    return result.code;
                }
                catch (e) {
                    this.logger.log('Error loading Svelte file:', path);
                    this.logger.debug('Error:', e);
                }
            }
            else {
                return readFile(path);
            }
        };
    }
}
exports.SvelteSnapshotManager = SvelteSnapshotManager;
